from dataclasses import dataclass, asdict, field
from typing import Dict


class ReverseMapping:
    """Class to provide reverse mapping functionality."""

    @property
    def reverse_mapping(self) -> Dict[int, str]:
        """Create a reverse mapping of the days."""
        return {v: k for k, v in asdict(self).items()}


@dataclass(frozen=True)
class DaysBookmark(ReverseMapping):
    Monday: int = 0
    Tuesday: int = 1
    Wednesday: int = 2
    Thursday: int = 3
    Friday: int = 4
    Saturday: int = 5
    Sunday: int = 6


@dataclass(frozen=True)
class EletTewsak(ReverseMapping):
    Saturday: int = 8
    Sunday: int = 7
    Monday: int = 6
    Tuesday: int = 5
    Wednesday: int = 4
    Thursday: int = 3
    Friday: int = 2


@dataclass(frozen=True)
class Wengelawyan(ReverseMapping):
    Matthew: int = 1
    Mark: int = 2
    Luke: int = 3
    John: int = 0


@dataclass(frozen=True)
class FastStartingDays:
    hudade: int = 14
    debrezeit: int = 41  # 11
    hosana: int = 62  # 2
    seklet: int = 67  # 7
    tensae: int = 69  # 9
    rekeb_kanat: int = 93  # 3
    erget: int = 108  # 18
    piraklitos: int = 118  # 28
    hawaryat: int = 119  # 29
    dehenet: int = 121  # 1


@dataclass(frozen=True)
class EthiopianCalendarMonths(ReverseMapping):
    Meskerem: int = 1
    Tikimt: int = 2
    Hidar: int = 3
    Tahsas: int = 4
    Tir: int = 5
    Yekatit: int = 6
    Megabit: int = 7
    Miyazia: int = 8
    Ginbot: int = 9
    Sene: int = 10
    Hamle: int = 11
    Nehasse: int = 12
    Pagumen: int = 13


@dataclass(frozen=True)
class MonthForFasting:
    hudade: Dict[str, int] = field(
        default_factory=lambda: {"Yekatit": 30, "Megabit": 60}
    )  # 6 7
    debrezeit: Dict[str, int] = field(
        default_factory=lambda: {"Yekatit": 30, "Megabit": 60, "Miyazia": 90}
    )  # x 6 7 8
    hosana: Dict[str, int] = field(
        default_factory=lambda: {"Megabit": 60, "Miyazia": 90}
    )  # 7 8
    seklet: Dict[str, int] = field(
        default_factory=lambda: {"Megabit": 60, "Miyazia": 90}
    )  # 7 8
    tensae: Dict[str, int] = field(
        default_factory=lambda: {"Megabit": 60, "Miyazia": 90}
    )  # 7 8
    rekeb_kanat: Dict[str, int] = field(
        default_factory=lambda: {"Miyazia": 90, "Ginbot": 120}
    )  # 8 9
    erget: Dict[str, int] = field(
        default_factory=lambda: {"Ginbot": 120, "Sene": 150}
    )  # 9 10
    piraklitos: Dict[str, int] = field(
        default_factory=lambda: {"Ginbot": 120, "Sene": 150}
    )  # 9 10
    hawaryat: Dict[str, int] = field(
        default_factory=lambda: {"Ginbot": 120, "Sene": 150}
    )  # 9 10
    dehenet: Dict[str, int] = field(
        default_factory=lambda: {"Ginbot": 120, "Sene": 150}
    )  # 9 10



from .lookups import FastStartingDays, MonthForFasting
from typing import Tuple


def add_days(day: str, num_days: int) -> str:
    """
    Adds the given number of days to the specified day of the week and returns the resulting day.

    Args:
        day (str): The day of the week (e.g., 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday').
        num_days (int): The number of days to add.

    Returns:
        str: The resulting day of the week.
    """
    days_of_the_week = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]
    day_index = days_of_the_week.index(day)
    new_index = (day_index + num_days) % len(days_of_the_week)
    return days_of_the_week[new_index]


def get_length_between(nenewe: int, fast: str) -> Tuple[int, int]:
    """
    Calculate the length between the nenewe day and the start of a fast.

    Args:
        nenewe (int): The nenewe day.
        fast (str): The name of the fast.

    Returns:
        Tuple[int, int]: The total length and the day of the event.
    """
    length_from_nenewe = getattr(FastStartingDays, fast) + nenewe
    day = 30 if (length_from_nenewe) % 30 == 0 else (length_from_nenewe) % 30

    return length_from_nenewe, day


def is_gregorian_leap_year(year: int) -> bool:
    """Check if the Gregorian year is a leap year."""
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def is_ethiopian_leap_year(year: int) -> bool:
    """Check if the Gregorian year is a leap year."""
    return year % 4 == 0


def calculate_gregorian_to_ethiopian(
    gregorian_year: int, gregorian_month: int, gregorian_day: int
) -> tuple[int, int, int]:
    """Convert a Gregorian date to an Ethiopian date.

    Args:
        gregorian_year (int): The Gregorian year.
        gregorian_month (int): The Gregorian month.
        gregorian_day (int): The Gregorian day.

    Returns:
        Tuple[int, int, int]: The Ethiopian year, month, and day.
    """
    if not 1 <= gregorian_year <= 9999:
        raise ValueError("Gregorian year must be a 4-digit integer")
    if not 1 <= gregorian_month <= 12:
        raise ValueError("Gregorian month must be between 1 and 12")

    # Calculate the number of days in the Gregorian month and year
    if gregorian_month in [1, 3, 5, 7, 8, 10, 12]:
        max_days = 31
    elif gregorian_month == 2:
        max_days = 28 if is_gregorian_leap_year(gregorian_year) else 29
    else:
        max_days = 30

    if not 1 <= gregorian_day <= max_days:
        raise ValueError(f"Gregorian day must be between 1 and {max_days}")

    # Difference in years between Gregorian and Ethiopian calendars
    ethio_year_diff = 7

    # New year date in Ethiopian calendar in Gregorian calendar
    if is_gregorian_leap_year(gregorian_year + 1):
        new_year_date = (9, 12)  # September 12 in a leap year
    else:
        new_year_date = (9, 11)  # September 11 in a common year

    # Calculate Ethiopian year
    diff_eight = False
    if (gregorian_month, gregorian_day) >= new_year_date:
        ethiopian_year = gregorian_year - ethio_year_diff
    else:
        ethiopian_year = gregorian_year - ethio_year_diff - 1
        diff_eight = True

    # Calculate the number of days since the last Ethiopian New Year
    gregorian_month_days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if is_gregorian_leap_year(gregorian_year):
        gregorian_month_days[1] = 29

    days_in_gregorian_year = sum(gregorian_month_days)
    days_since_new_year = (
        sum(gregorian_month_days[: gregorian_month - 1]) + gregorian_day
    )
    days_since_new_year -= (
        sum(gregorian_month_days[: new_year_date[0] - 1]) + new_year_date[1]
    )

    if days_since_new_year < 0:
        days_since_new_year += days_in_gregorian_year

    # Calculate Ethiopian month and day
    ethiopian_month = days_since_new_year // 30 + 1
    # ethiopian_day = days_since_new_year % 30 + 1
    if diff_eight:
        ethiopian_day = days_since_new_year % 30
    else:
        ethiopian_day = days_since_new_year % 30 + 1
    if ethiopian_month == 13 and ethiopian_day < 5:
        ethiopian_day += 1

    if ((ethiopian_year + 1) % 4) == 0 and (gregorian_month, gregorian_day) == (
        new_year_date[0],
        new_year_date[1] - 1,
    ):
        return ethiopian_year, 13, 6

    return ethiopian_year, ethiopian_month, ethiopian_day


def calculate_ethiopian_to_gregorian(
    ethiopian_year: int, ethiopian_month: int, ethiopian_day: int
) -> tuple[int, int, int]:
    """Convert an Ethiopian date to a Gregorian date.

    Args:
        ethiopian_year (int): The Ethiopian year.
        ethiopian_month (int): The Ethiopian month.
        ethiopian_day (int): The Ethiopian day.

    Returns:
        Tuple[int, int, int]: The Gregorian year, month, and day.
    """
    if not 1 <= ethiopian_year <= 9999:
        raise ValueError("Ethiopian year must be a 4-digit integer")
    if not 1 <= ethiopian_month <= 13:
        raise ValueError("Ethiopian month must be between 1 and 13")
    if not 1 <= ethiopian_day <= 30:
        raise ValueError("Ethiopian day must be between 1 and 30")

    # Calculate the number of days in the Ethiopian month
    max_days = 30
    if ethiopian_month == 13:
        max_days = 6 if (ethiopian_year % 4) == 3 else 5

    if not 1 <= ethiopian_day <= max_days:
        raise ValueError(f"Ethiopian day must be between 1 and {max_days}")

    # Difference in years between Gregorian and Ethiopian calendars
    ethio_year_diff = 7

    gregorian_year = ethiopian_year + ethio_year_diff

    # New year date in Ethiopian calendar in Gregorian calendar
    if is_gregorian_leap_year(gregorian_year + 1):
        new_year_date = (9, 12)  # September 12 in a leap year
    else:
        new_year_date = (9, 11)  # September 11 in a common year

    # Calculate the number of days since the Ethiopian New Year
    days_since_new_year = (ethiopian_month - 1) * 30 + (ethiopian_day - 1)

    # Calculate the Gregorian date
    gregorian_month_days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if is_gregorian_leap_year(gregorian_year + 1):
        gregorian_month_days[1] = 29

    days_in_gregorian_year = sum(gregorian_month_days)
    new_year_day_of_year = (
        sum(gregorian_month_days[: new_year_date[0] - 1]) + new_year_date[1]
    )

    gregorian_day_of_year = new_year_day_of_year + days_since_new_year
    if gregorian_year - ethiopian_year == 8:
        gregorian_day_of_year -= 1

    if gregorian_day_of_year > days_in_gregorian_year:
        gregorian_day_of_year -= days_in_gregorian_year
        gregorian_year += 1

    # Calculate the month and day
    gregorian_month = 1
    while gregorian_day_of_year > gregorian_month_days[gregorian_month - 1]:
        gregorian_day_of_year -= gregorian_month_days[gregorian_month - 1]
        gregorian_month += 1

    gregorian_day = gregorian_day_of_year

    if ((ethiopian_year + 1) % 4) == 0 and (ethiopian_month, ethiopian_day) == (13, 6):
        return gregorian_year, 9, 11

    return gregorian_year, gregorian_month, gregorian_day


# @staticmethod
def get_total_days(total_years: int) -> int:
    leap_years = total_years // 4
    return (total_years - 1) * 365 + leap_years


# @staticmethod
def get_day_of_week(total_days: int) -> int:
    return (total_days % 7) + 1


# @staticmethod
def calculate_days_to_nenewe(start_day_of_week: int) -> int:
    return 128 - ((start_day_of_week + 1) % 7) + 1


# @staticmethod
def calculate_event_date(
    total_days_till_metke: int,
    days_to_event: int,
    nenewe_to_event_length: int,
    total_days: int,
) -> Tuple[int, int]:
    event_day = total_days_till_metke + days_to_event + nenewe_to_event_length
    month_of_event = ((event_day - total_days - 1) // 30) + 1
    if (event_day - total_days - 1) % 30 == 0:
        month_of_event -= 1
    day_of_event = ((event_day - total_days - 1) % 30) or 30
    return month_of_event, day_of_event


def old_get_event_date(self, event_name: str) -> str:
    """
    Calculate the date of an event based on the nenewe day and the length to the event.

    Args:
        event_name (str): The name of the event.

    Returns:
        str: The date of the event.
    """
    nenewe, metke_month = self.get_nenewe()
    length_to_event, day_of_event = self.get_length_between(nenewe, event_name)

    months = MonthForFasting()
    months = getattr(months, event_name)

    # min_value = min(months.values())

    for month, value in months.items():
        if metke_month > 0:
            if length_to_event > value:
                continue
        else:
            if length_to_event - 30 > value:
                continue
        return f"{month} {day_of_event}"



from .lookups import (
    DaysBookmark,
    EletTewsak,
    FastStartingDays,
    Wengelawyan,
    EthiopianCalendarMonths,
)
from .helper import (
    add_days,
    get_total_days,
    get_day_of_week,
    calculate_days_to_nenewe,
    calculate_event_date,
    calculate_ethiopian_to_gregorian,
    calculate_gregorian_to_ethiopian,
)
from typing import Tuple


class BaherHasab:
    def __init__(self, given_year: int = 2016) -> None:
        """
        Initialize the BaherHasab class with the given current year.

        Args:
            given_year (int): The current year in the Ethiopian calendar (its 7 or 8 years less than the
                                    Geogorian calendar depending on the season of the year).
        """
        self.abiy_kemer = 532  # also known as abiy awde
        self.awde_mahtot = 76  # also known as mahkelawi awde
        self.awde_tsehay = 28  # also known as nuhase awde
        self.awde_abketa = 19
        self.old_abketa = 11
        self.amet_alem = 5500
        self.given_year = given_year

    def get_total_years(self) -> int:
        """
        Calculate the total years since the creation of the world according to the Ethiopian calendar.

        Returns:
            int: The total number of years.
        """
        return self.amet_alem + self.given_year

    def get_wember(self) -> int:
        """
        Calculate the Wember value for the current year.
        (we minus one because the year has started not finished)

        Returns:
            int: The Wember value.
        """
        total_years = self.get_total_years()
        return (total_years - 1) % self.awde_abketa

    def get_abketa(self) -> int:
        """
        Calculate the Abketa value based on the Wember value.

        Returns:
            int: The Abketa value.
        """
        wember = self.get_wember()
        abketa = (wember * 11) % 30
        return abketa

    def get_metke(self) -> int:
        """
        Calculate the Metke value based on the Abketa value.

        Returns:
            int: The Metke value.
        """
        abketa = self.get_abketa()
        return 30 - abketa

    def get_first_day_of_year(self) -> str:
        """
        Calculate the first day of the year according to the Ethiopian calendar.

        Returns:
            str: The first day of the year.
        """
        total_years = self.get_total_years()
        first_day = (total_years + (total_years // 4)) % 7
        return DaysBookmark().reverse_mapping[first_day]

    def get_nenewe(self) -> Tuple[int, int]:
        """
        Calculate the nenewe day and Metke month based on the Metke value.

        Returns:
            Tuple[int, int]: The nenewe day and Metke month.
        """
        metke = self.get_metke()
        metke_month = 0 if metke > 13 else 30
        first_day = self.get_first_day_of_year()
        day_of_metke = add_days(first_day, metke_month + metke - 1)
        day_of_nenewe = getattr(EletTewsak, day_of_metke)

        nenewe_day = (metke_month + metke + day_of_nenewe) % 30

        print(
            f"Nenewe Starts at Tir {nenewe_day}"
            if day_of_nenewe + metke <= 30 and metke_month == 0
            else f"Nenewe Starts at Yekatit {nenewe_day}"
        )

        return nenewe_day, metke_month

    def get_event_date(self, event_name: str) -> str:
        """
        Calculate the date of an event based on the nenewe day and the length to the event.

        Args:
            event_name (str): The name of the event.

        Returns:
            str: The date of the event.
        """

        total_years = self.get_total_years()
        metke = self.get_metke()
        nenewe_day, metke_month = self.get_nenewe()

        # Total days from all years including B.C and A.D
        total_days = get_total_days(total_years)
        # Total days from all years including B.C and A.D till the metke of the year
        total_days_till_metke = total_days + metke_month + metke

        day_of_week_of_metke = get_day_of_week(total_days_till_metke)
        # Getting the Twesak of each event
        nenewe_to_event_length = getattr(FastStartingDays(), event_name)

        days_to_nenewe = calculate_days_to_nenewe(day_of_week_of_metke)
        month_of_event, day_of_event = calculate_event_date(
            total_days_till_metke, days_to_nenewe, nenewe_to_event_length, total_days
        )

        return f"{EthiopianCalendarMonths().reverse_mapping[month_of_event]} {day_of_event}"

    def get_hudade(self) -> str:
        """
        Get the date of Hudade.

        Returns:
            str: The date of Hudade.
        """
        return self.get_event_date("hudade")

    def get_debrezeit(self) -> str:
        """
        Get the date of Debre Zeit.

        Returns:
            str: The date of Debre Zeit.
        """
        return self.get_event_date("debrezeit")

    def get_hosana(self) -> str:
        """
        Get the date of Hosana.

        Returns:
            str: The date of Hosana.
        """
        return self.get_event_date("hosana")

    def get_seklet(self) -> str:
        """
        Get the date of Seklet.

        Returns:
            str: The date of Seklet.
        """
        return self.get_event_date("seklet")

    def get_tensae(self) -> str:
        """
        Get the date of Tensae.

        Returns:
            str: The date of Tensae.
        """
        return self.get_event_date("tensae")

    def get_rekeb_kanat(self) -> str:
        """
        Get the date of Rekeb Kanat.

        Returns:
            str: The date of Rekeb Kanat.
        """
        return self.get_event_date("rekeb_kanat")

    def get_erget(self) -> str:
        """
        Get the date of Erget.

        Returns:
            str: The date of Erget holiday.
        """
        return self.get_event_date("erget")

    def get_piraklitos(self) -> str:
        """
        Get the date of Piraklitos.

        Returns:
            str: The date of Piraklitos holiday.
        """
        return self.get_event_date("piraklitos")

    def get_hawaryat(self) -> str:
        """
        Get the date of Hawaryat.

        Returns:
            str: The date of Hawaryat fast.
        """
        return self.get_event_date("hawaryat")

    def get_dehenet(self) -> str:
        """
        Get the date of Dehenet.

        Returns:
            str: The date of Dehenet fast.
        """
        return self.get_event_date("dehenet")

    def get_wengelawyan(self) -> str:
        """
        Get the gospel of the year.

        Returns:
            str: The Gospel (Wengelawyan) of the year
        """
        gospels = self.given_year % 4
        return Wengelawyan().reverse_mapping[gospels]

    def get_awde_kemer(self) -> Tuple[int, int, int]:
        """
        Get the awde kemer.

        Returns:
            int: the ith awde kemer we are at (based on the given year)
            int: the past years in the awde kemer
            int: the years left for the awde kemer to finish
        """
        total_years = self.get_total_years()
        ith_abiy_kemer = (total_years // self.abiy_kemer) + 1
        passed_years = total_years % self.abiy_kemer
        reminded_years = self.abiy_kemer - passed_years
        print(f"""We are in the {ith_abiy_kemer}th Abiy kemer \n 
              have gone through {passed_years} years \n
              and have {reminded_years} years left""")
        return ith_abiy_kemer, passed_years, reminded_years

    def get_awde_mahtot(self) -> Tuple[int, int, int]:
        """
        Get the awde mahtot.

        Returns:
            int: the ith awde mahtot we are at (based on the given year)
            int: the past years in the awde mahtot
            int: the years left for the awde mahtot to finish
        """
        total_years = self.get_total_years()
        ith_awde_mahtot = (total_years // self.awde_mahtot) + 1
        passed_years = total_years % self.awde_mahtot
        reminded_years = self.awde_mahtot - passed_years
        print(f"""We are in the {ith_awde_mahtot}th Awde Mahtot \n 
              have gone through {passed_years} years \n
              and have {reminded_years} years left""")
        return ith_awde_mahtot, passed_years, reminded_years

    def get_awde_tsehay(self) -> Tuple[int, int, int]:
        """
        Get the awde tsehay.

        Returns:
            int: the ith awde tsehay we are at (based on the given year)
            int: the past years in the awde tsehay
            int: the years left for the awde tsehay to finish
        """
        total_years = self.get_total_years()
        ith_awde_tsehay = (total_years // self.awde_tsehay) + 1
        passed_years = total_years % self.awde_tsehay
        reminded_years = self.awde_tsehay - passed_years
        print(f"""We are in the {ith_awde_tsehay}th Awde Mahtot \n 
              have gone through {passed_years} years \n
              and have {reminded_years} years left""")
        return ith_awde_tsehay, passed_years, reminded_years

    def ethiopian_to_gregorian(
        ethiopian_year: int, ethiopian_month: int, ethiopian_day: int
    ) -> tuple[int, int, int]:
        """Convert an Ethiopian date to a Gregorian date.

        Args:
            ethiopian_year (int): The Ethiopian year.
            ethiopian_month (int): The Ethiopian month.
            ethiopian_day (int): The Ethiopian day.

        Returns:
            Tuple[int, int, int]: The Gregorian year, month, and day.
        """

        return calculate_ethiopian_to_gregorian(
            ethiopian_year, ethiopian_month, ethiopian_day
        )

    def gregorian_to_ethiopian(
        gregorian_year: int, gregorian_month: int, gregorian_day: int
    ) -> tuple[int, int, int]:
        """Convert a Gregorian date to an Ethiopian date.

        Args:
            gregorian_year (int): The Gregorian year.
            gregorian_month (int): The Gregorian month.
            gregorian_day (int): The Gregorian day.

        Returns:
            Tuple[int, int, int]: The Ethiopian year, month, and day.
        """
        return calculate_gregorian_to_ethiopian(
            gregorian_year, gregorian_month, gregorian_day
        )
